(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

theory Completion_Fairness
imports
  Abstract_Completion
  Prime_Critical_Pairs
begin

locale kb_inf =
  kb less for less :: "('a, 'b :: infinite) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
begin

lemma CP_subset_imp_rstep_CP_subset:
  assumes "CP R \<subseteq> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (CP R))\<^sup>\<leftrightarrow> \<subseteq> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)" (is "_ \<subseteq> ?A")
proof
  fix s t
  assume "(s, t) \<in> (rstep (CP R))\<^sup>\<leftrightarrow>"
  then have "(s, t) \<in> rstep (CP R) \<or> (t, s) \<in> rstep (CP R)" by auto
  moreover
  { fix s t
    assume "(s, t) \<in> rstep (CP R)"
    then have "(s, t) \<in> ?A"
    proof (rule rstepE)
      fix C \<sigma> l r
      assume cp: "(l, r) \<in> CP R"
        and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      from assms [THEN subsetD, OF cp]
        show ?thesis by auto
    qed }
  ultimately show "(s, t) \<in> ?A" by (metis (no_types) Union_sym)
qed

lemma finite_run_CP:
  assumes "R 0 = {}" and "E n = {}"
    and run: "\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))"
    and fair: "CP (R n) \<subseteq> (\<Union>i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>* \<and> SN (rstep (R n)) \<and> CR (rstep (R n))"
proof -
  have R_less: "\<And>i. i \<le> n \<Longrightarrow> R i \<subseteq> {(x, y). x \<succ> y}"
    and "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>*"
    and "SN (rstep (R n))"
    by (fact finite_runD [OF assms(1-3)])+
  note vc =  SN_imp_variable_condition [OF this(3)]
  from peak_imp_join_or_CP [OF vc]
    and CP_subset_imp_rstep_CP_subset [OF fair]
    have "(rstep (R n))\<inverse> O rstep (R n) \<subseteq> (rstep (R n))\<^sup>\<down> \<union> (\<Union> i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)"
    by (blast elim: peak_iff [THEN iffD1, of _ _ "rstep (R n)" "rstep (R n)"])
  from finite_run [OF assms(1-3) this]
    show ?thesis .
qed

lemma PCP_subset_imp_rstep_PCP_subset:
  assumes "PCP R \<subseteq> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (PCP R))\<^sup>\<leftrightarrow> \<subseteq> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)" (is "_ \<subseteq> ?A")
proof
  fix s t
  assume "(s, t) \<in> (rstep (PCP R))\<^sup>\<leftrightarrow>"
  then have "(s, t) \<in> rstep (PCP R) \<or> (t, s) \<in> rstep (PCP R)" by auto
  moreover
  { fix s t
    assume "(s, t) \<in> rstep (PCP R)"
    then have "(s, t) \<in> ?A"
    proof (rule rstepE)
      fix C \<sigma> l r
      assume pcp: "(l, r) \<in> PCP R"
        and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      from assms [THEN subsetD, OF pcp]
        show ?thesis by auto
    qed }
  ultimately show "(s, t) \<in> ?A" by (metis (no_types) Union_sym)
qed

lemma PCP_subset_imp_rstep_PCP_subset':
  assumes "PCP R \<subseteq> (rstep R)\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (PCP R))\<^sup>\<leftrightarrow> \<subseteq> (rstep R)\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)" (is "_ \<subseteq> ?A")
proof
  fix s t
  assume "(s, t) \<in> (rstep (PCP R))\<^sup>\<leftrightarrow>"
  then have *: "(s, t) \<in> rstep (PCP R) \<or> (t, s) \<in> rstep (PCP R)" by auto
  moreover
  { fix s t
    assume "(s, t) \<in> rstep (PCP R)"
    then have "(s, t) \<in> ?A"
    proof (rule rstepE)
      fix C \<sigma> l r
      assume pcp: "(l, r) \<in> PCP R"
        and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      from assms [THEN subsetD, OF pcp]
        show ?thesis by(auto) (metis atMost_iff rstep.simps)
    qed }
  ultimately
    show "(s, t) \<in> ?A" by (auto simp del: reverseTrs intro: join_sym)
qed

lemma fair_imp_nabla:
  assumes "PCP R \<subseteq> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)"
  shows "\<And>s t u. nabla R s t u \<Longrightarrow> (t, u) \<in> (rstep R)\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)"
  using PCP_subset_imp_rstep_PCP_subset [OF assms]
  by (auto simp add: nabla_def)

lemma fair_new_imp_nabla:
  assumes "PCP R \<subseteq> (rstep R)\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)"
  shows "\<And>s t u. nabla R s t u \<Longrightarrow> (t, u) \<in> (rstep R)\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (S i))\<^sup>\<leftrightarrow>)"
  using PCP_subset_imp_rstep_PCP_subset' [OF assms]
  by (auto simp: nabla_def)

lemma finite_run':
  assumes "R 0 = {}" and "E n = {}"
    and run: "\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))"
    and nabla: "\<And>s t u. nabla (R n) s t u \<Longrightarrow> (t, u) \<in> (rstep (R n))\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>* \<and> SN (rstep (R n)) \<and> CR (rstep (R n))"
proof -
  have R_less: "\<And>i. i \<le> n \<Longrightarrow> R i \<subseteq> {(x, y). x \<succ> y}"
    and "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>*"
    and SN: "SN (rstep (R n))"
    by (fact finite_runD [OF assms(1-3)])+
  moreover
  have "CR (rstep (R n))"
  proof -
    def \<R> \<equiv> "\<Union>M\<in>UNIV. mstep M (R n)"
    def mless \<equiv> "\<lambda>x y. mulex less\<inverse>\<inverse> y x"
    write mless (infix "\<succ>''" 50)
    interpret lab: ars_labeled_sn \<R> "\<lambda>M. mstep M (R n)" UNIV "op \<succ>'"
    proof
      show "\<R> = (\<Union>M. mstep M (R n))" using \<R>_def R_less by simp
      show "SN {(x, y). x \<succ>' y}" unfolding mless_def by (metis SN_less SN_mulex)
    qed

    interpret ars_peak_decreasing \<R> "\<lambda>M. mstep M (R n)" UNIV "op \<succ>'"
    proof
      fix s t u :: "('a, 'b) term" and M\<^sub>1 M\<^sub>2 :: "('a, 'b) term multiset"
      def M \<equiv> "{#t, u#}"
      assume "(s, t) \<in> mstep M\<^sub>1 (R n)" and "(s, u) \<in> mstep M\<^sub>2 (R n)"
      then have *: "(s, t) \<in> (rstep (R n))" "(s, u) \<in> rstep (R n)" by (auto iff: mstep_iff)
      { fix v w
        assume "nabla (R n) s v w"
        then have "(s, v) \<in> (rstep (R n))\<^sup>+" and "(s, w) \<in> (rstep (R n))\<^sup>+" by (auto simp: nabla_def)
        have "M\<^sub>1 \<succ>' {#v, w#}"
        proof -
          from \<open>(s, t) \<in> mstep M\<^sub>1 (R n)\<close> have ne1: "M\<^sub>1 \<noteq> {#}" by (auto iff: mstep_iff)
          have "\<forall>u. u \<in># {#v, w#} \<longrightarrow> (\<exists>z. z \<in># M\<^sub>1 \<and> z \<succ> u)"
          proof -
            { have "\<exists>z. z \<in># M\<^sub>1 \<and> z \<succ> v"
              using \<open>(s, t) \<in> mstep M\<^sub>1 (R n)\<close> and \<open>(s, v) \<in> (rstep (R n))\<^sup>+\<close> 
              and rsteps_subset_less [OF R_less [of n]] by (auto simp: mstep_iff dest: trans) }
            moreover
            { have "\<exists>z. z \<in># M\<^sub>1 \<and> z \<succ> w"
              using \<open>(s, t) \<in> mstep M\<^sub>1 (R n)\<close> and \<open>(s, w) \<in> (rstep (R n))\<^sup>+\<close>
              and rsteps_subset_less [OF R_less [of n]] by (auto simp: mstep_iff dest: trans) }
            ultimately show ?thesis by (auto simp: M_def)
          qed
          from mulex_on_all_strict [OF ne1 _ _ this, of UNIV]
            have less1: "mulex less\<inverse>\<inverse> {#v, w#} M\<^sub>1" by auto (metis (lifting, no_types) conversep.intros mulex_on_mono)
          then show ?thesis by (simp add: mless_def)
        qed }
      note nabla_less = this

      let ?D = "(\<Union>c\<in>lab.downset2 M\<^sub>1 M\<^sub>2. mstep c (R n))\<^sup>\<leftrightarrow>\<^sup>*"

      { fix v w
        assume nabla': "nabla (R n) s v w"
        let ?M = "{#v, w#}"
        from nabla [OF nabla']
          have "(v, w) \<in> (rstep (R n))\<^sup>\<down> \<union> (\<Union> i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)" .
        then have "(v, w) \<in> ?D"
        proof
          assume "(v, w) \<in> (rstep (R n))\<^sup>\<down>"
          then obtain u where 1: "(v, u) \<in> (rstep (R n))\<^sup>*"
            and 2: "(w, u) \<in> (rstep (R n))\<^sup>*" by auto
          from rsteps_imp_msteps [of ?M, OF _ 1 R_less]
            have "(v, u) \<in> (mstep ?M (R n))\<^sup>\<leftrightarrow>\<^sup>*" by (auto simp: M_def)
          from lab.conversion_in_downset2 [OF this] and nabla_less [OF nabla']
            have 3: "(v, u) \<in> ?D" by simp
          from rsteps_imp_msteps [of ?M, OF _ 2 R_less]
            have "(u, w) \<in> (mstep ?M (R n))\<^sup>\<leftrightarrow>\<^sup>*" by (auto simp: M_def) (metis conversion_def in_rtrancl_UnI rtrancl_converseI)
          from lab.conversion_in_downset2 [OF this] and nabla_less [OF nabla']
            have 4: "(u, w) \<in> ?D" by simp
          from 3 and 4 show ?thesis by (auto simp: conversion_def)
        next
          assume "(v, w) \<in> (\<Union>i \<le> n. (rstep (E i))\<^sup>\<leftrightarrow>)"
          then obtain i where "i \<le> n" and "(v, w) \<in> (rstep (E i))\<^sup>\<leftrightarrow>" by auto
          then have "(v, w) \<in> (mstep ?M (E i))\<^sup>\<leftrightarrow>" by (auto iff: mstep_iff simp: M_def)
          then have "(v, w) \<in> (mstep ?M (E i \<union> R i))\<^sup>\<leftrightarrow>\<^sup>*" by (auto simp: conversion_def)
          then show ?thesis using \<open>i \<le> n\<close>
          proof (induct "n - i" arbitrary: i)
            case 0
            with \<open>i \<le> n\<close> have [simp]: "i = n" by arith
            with 0 have "(v, w) \<in> (mstep ?M (R n))\<^sup>\<leftrightarrow>\<^sup>*" by (simp add: assms)
            from lab.conversion_in_downset2 [OF this] and nabla_less [OF nabla']
              show ?case by auto
          next
            case (Suc m)
            from \<open>Suc m = n - i\<close> have *: "m = n - Suc i" and leq: "Suc i \<le> n" by auto
            from run have **: "(E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))" using leq by auto
            from \<open>(v, w) \<in> (mstep ?M (E i \<union> R i))\<^sup>\<leftrightarrow>\<^sup>*\<close>
              have "(v, w) \<in> (mstep ?M (E (Suc i) \<union> R (Suc i)))\<^sup>\<leftrightarrow>\<^sup>*"
              using msteps_subset [OF ** R_less [OF leq]] by blast
            from Suc.hyps(1) [OF * this leq]
              show ?case .
          qed
        qed }
      note ** = this

      from peak_imp_nabla2 [OF SN_imp_variable_condition [OF SN] *]
        show "(t, u) \<in> ?D"
        by (auto dest!: ** simp: conversion_def)
    qed
    from CR show "CR (rstep (R n))" by (simp add: \<R>_def)
  qed
  ultimately show ?thesis by blast
qed

text \<open>Every finite and fair run of abstract completion yields a complete TRS
for the given set of equations.\<close>
lemma finite_fair_run:
  assumes "R 0 = {}" and "E n = {}"
    and "\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))"
    and fair: "PCP (R n) \<subseteq> (\<Union>i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>* \<and> SN (rstep (R n)) \<and> CR (rstep (R n))"
  using finite_run' [OF assms(1-3)] and fair_imp_nabla [OF fair] by blast

lemma finite_fair_new_run:
  assumes "R 0 = {}" and "E n = {}"
    and "\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))"
    and fair: "PCP (R n) \<subseteq> (rstep (R n))\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>* \<and> SN (rstep (R n)) \<and> CR (rstep (R n))"
  using finite_run' [OF assms(1-3)] and fair_new_imp_nabla [OF fair] by blast

end

end

